import logging as log
from typing import Type

from PyQt5.QtCore import QAbstractTableModel, Qt, QModelIndex, QVariant
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QTableView

from hspylib.core.tools.commons import class_attribute_values, class_attribute_names


class DefaultTableModel(QAbstractTableModel):
    def __init__(self, clazz: Type, headers: list = None, cell_alignments: list = None, table_data: list = None,
                 parent: QTableView = None):
        QAbstractTableModel.__init__(self, parent)
        self.clazz = clazz
        self.table_data = table_data or []
        self.headers = headers or self.headers_by_entity()
        self.cell_alignments = cell_alignments or []
        log.info('{} table_headers={}'.format(clazz.__class__.__name__, '|'.join(self.headers)))

    def data(self, index: QModelIndex, role: int = ...):
        entity = class_attribute_values(self.table_data[index.row()].__dict__)[index.column()]
        str_entity = str(entity) if entity else ''
        if role == Qt.DisplayRole:
            return str_entity
        elif role == Qt.TextAlignmentRole:
            return self.cell_alignments[index.column()] if self.cell_alignments else Qt.AlignLeft
        elif role == Qt.BackgroundColorRole:
            return QVariant() if entity else QColor(230, 230, 230)

        return QVariant()

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.headers[section].upper() if len(self.headers) >= section else '-'
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            return section
        return QVariant()

    def headers_by_entity(self):
        return class_attribute_names(self.clazz)

    def rowCount(self, parent: QModelIndex = ...):
        return len(self.table_data) if self.table_data and len(self.table_data) > 0 else 0

    def columnCount(self, parent: QModelIndex = ...):
        return len(self.table_data[0].__dict__.keys()) if self.table_data and len(self.table_data) > 0 else 0

    def row(self, index: QModelIndex):
        return self.table_data[index.row()]



from abc import ABC, abstractmethod
from typing import Optional, List

from requests.structures import CaseInsensitiveDict

from hspylib.core.model.entity import Entity


class CrudRepository(ABC):
    @abstractmethod
    def insert(self, entity: Entity):
        pass

    @abstractmethod
    def update(self, entity: Entity):
        pass

    @abstractmethod
    def delete(self, entity: Entity):
        pass

    @abstractmethod
    def find_all(self, filters: CaseInsensitiveDict = None) -> List[Entity]:
        pass

    @abstractmethod
    def find_by_id(self, entity_id: str) -> Optional[Entity]:
        pass



import os
import logging as log
from typing import List, Optional

from requests.structures import CaseInsensitiveDict

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.meta.singleton import Singleton
from hspylib.core.model.entity import Entity

DEFAULT_SQL_STUBS = '{}/sql/sql_stubs.sql'.format(os.path.dirname(__file__))


class SqlFactory(metaclass=Singleton):

    @staticmethod
    def read_stubs(sql_filename: str) -> dict:
        ret_val = {}
        assert os.path.exists(sql_filename), "Sql file was not found: {}".format(sql_filename)
        with open(sql_filename) as f_stubs:
            lines = f_stubs.readlines()
            assert lines, "Stub file is empty"
            lines = list(map(str.strip, lines))
            stubs = ' '.join(lines).split(';')
            assert len(stubs) >= 4, "Stub file hasn't got the minimum stubs for [insert, select, update, delete]"
            for stub in stubs:
                if stub:
                    key = stub.strip().partition(' ')[0].lower()
                    ret_val[key] = stub.strip()
        return ret_val

    @staticmethod
    def join_filters(filters: CaseInsensitiveDict, join_operator: str = 'AND') -> str:
        filter_string = ''
        if filters:
            for key, value in filters.items():
                filter_string += "{} {} = '{}'".format(join_operator, key, value)
        return filter_string

    @staticmethod
    def join_fieldset(entity: Entity) -> str:
        fields = entity.to_column_set()
        field_set = ''
        for key, value in fields.items():
            field_set += "{}{} = '{}'".format(', ' if field_set else '', key, value)
        return field_set

    def __init__(self):
        self.logger = AppConfigs.INSTANCE.logger()
        self.sql_stubs = SqlFactory.read_stubs(DEFAULT_SQL_STUBS)
        log.debug('{} created with {} Stubs'.format(
            self.__class__.__name__,
            len(self.sql_stubs)))

    def insert(self, entity: Entity) -> Optional[str]:
        params = entity.to_values()
        sql = self.sql_stubs['insert']\
                  .replace(':columnSet', str(entity.to_columns()).replace("'", ""))\
                  .replace(':valueSet', str(params))
        return sql

    def select(self, column_set: List[str] = None, filters: CaseInsensitiveDict = None) -> Optional[str]:
        sql = self.sql_stubs['select']\
            .replace(':columnSet', '*' if not column_set else ', '.join(column_set))\
            .replace(':filters', SqlFactory.join_filters(filters))
        return sql

    def update(self, entity: Entity, filters: CaseInsensitiveDict) -> Optional[str]:
        sql = self.sql_stubs['update']\
            .replace(':fieldSet', SqlFactory.join_fieldset(entity))\
            .replace(':filters', SqlFactory.join_filters(filters))
        return sql

    def delete(self, filters: CaseInsensitiveDict) -> Optional[str]:
        sql = self.sql_stubs['delete']\
            .replace(':filters', SqlFactory.join_filters(filters))
        return sql



import json
import logging as log
import uuid
from abc import abstractmethod
from typing import Optional

from requests.exceptions import HTTPError
from requests.structures import CaseInsensitiveDict

from hspylib.core.crud.crud_repository import CrudRepository
from hspylib.core.crud.db.firebase.firebase_config import FirebaseConfig
from hspylib.core.enum.http_code import HttpCode
from hspylib.core.model.entity import Entity
from hspylib.modules.fetch.fetch import put, get, delete


class FirebaseRepository(CrudRepository):
    def __init__(self):
        self.payload = None
        self.config = FirebaseConfig()

    def __str__(self):
        return str(self.payload)

    def to_list(self, json_string: str, filters: CaseInsensitiveDict = None) -> list:
        the_list = []
        for key, value in json.loads(json_string).items():
            if not filters or all(k in value and value[k] == v for k, v in filters.items()):
                the_list.append(self.row_to_entity(value))
        return the_list

    def insert(self, entity: Entity):
        entity.uuid = entity.uuid if entity.uuid is not None else str(uuid.uuid4())
        url = '{}/{}.json'.format(self.config.base_url(), entity.uuid)
        payload = entity.to_json()
        log.debug('Inserting firebase entry: {} into: {}'.format(entity, url))
        response = put(url, payload)
        assert response, "Response is empty"
        if response.status_code != HttpCode.OK:
            raise HTTPError('{} - Unable to put into={} with json_string={}'.format(response.status_code, url, payload))

    def update(self, entity: Entity):
        url = '{}/{}.json'.format(self.config.base_url(), entity.uuid)
        payload = entity.to_json()
        log.debug('Updating firebase entry: {} into: {}'.format(entity, url))
        response = put(url, payload)
        assert response, "Response is empty"
        if response.status_code != HttpCode.OK:
            raise HTTPError('{} - Unable to put into={} with json_string={}'.format(response.status_code, url, payload))

    def delete(self, entity: Entity):
        url = '{}/{}.json'.format(self.config.base_url(), entity.uuid)
        log.debug('Deleting firebase entry: {} into: {}'.format(entity, url))
        response = delete(url)
        assert response, "Response is empty"
        if response.status_code != HttpCode.OK:
            raise HTTPError('{} - Unable to delete from={}'.format(response.status_code, url))

    def find_all(self, filters: CaseInsensitiveDict = None) -> Optional[list]:
        url = '{}.json?orderBy="$key"'.format(self.config.base_url())
        log.debug('Fetching firebase entries from {}'.format(url))
        response = get(url)
        assert response, "Response is empty"
        if response.status_code != HttpCode.OK:
            raise HTTPError('{} - Unable to get from={}'.format(response.status_code, url))

        return self.to_list(response.body, filters) if response.body else []

    def find_by_id(self, entity_id: str) -> Optional[Entity]:
        url = '{}.json?orderBy="$key"&equalTo="{}"'.format(self.config.base_url(), entity_id)
        log.debug('Fetching firebase entry entity_id={} from {}'.format(entity_id, url))
        response = get(url)
        assert response, "Response is empty"
        if response.status_code != HttpCode.OK:
            raise HTTPError('{} - Unable to get from={}'.format(response.status_code, url))
        result = self.to_list(response.body) if response.body else []
        assert len(result) <= 1, "Multiple results found with entity_id={}".format(entity_id)

        return result[0] if len(result) > 0 else None

    @abstractmethod
    def row_to_entity(self, row: dict) -> Entity:
        pass

    @abstractmethod
    def database_name(self) -> str:
        pass



import json
from typing import Tuple
from uuid import UUID


class Entity(object):
    def __init__(self, entity_id: UUID = None):
        self.uuid = entity_id

    def __str__(self):
        return "{}:{}".format(
            self.__class__.__name__,
            str(self.to_dict()))

    def __eq__(self, other):
        return isinstance(other, Entity) and all(item in self.to_dict().items() for item in other.to_dict().items())

    def to_dict(self) -> dict:
        ret_dict = {}
        for key, value in self.__dict__.items():
            if isinstance(value, bool):
                ret_dict[key] = bool(value)
            elif isinstance(value, int):
                ret_dict[key] = int(value)
            elif isinstance(value, float):
                ret_dict[key] = float(value)
            elif isinstance(value, str) or isinstance(value, UUID):
                ret_dict[key] = str(value)
            else:
                ret_dict[key] = value.__dict__() if value else {}
        return ret_dict

    def to_json(self):
        dict_obj = self.to_dict()
        json_str = json.dumps(dict_obj)
        return json_str

    def to_columns(self) -> Tuple[str]:
        cols = []
        for key in self.__dict__.keys():
            if not key.startswith('_'):
                cols.append(key.replace("'", "").upper())
        return tuple(cols)

    def to_column_set(self) -> dict:
        fields = {}
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                fields[key.replace("'", "").upper()] = "{}".format(str(value))
        return fields

    def to_values(self) -> Tuple[str]:
        values = []
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                values.append(str(value))
        return tuple(values)



#!/usr/bin/env python3
import re
import signal
from abc import ABC
from typing import List, Any

from hspylib.core.tools.commons import sysout, screen_size
from hspylib.core.tools.keyboard import Keyboard
from hspylib.ui.cli.menu.menu_utils import MenuUtils
from hspylib.ui.cli.vt100.vt_100 import Vt100
from hspylib.ui.cli.vt100.vt_codes import vt_print
from hspylib.ui.cli.vt100.vt_colors import VtColors


def mchoose(
        items: List[Any],
        checked: bool = True,
        title: str = 'Please select one',
        max_rows: int = 15,
        title_color: VtColors = VtColors.ORANGE,
        highlight_color: VtColors = VtColors.BLUE,
        nav_color: VtColors = VtColors.YELLOW) -> List[Any]:
    """
    TODO
    :param items:
    :param checked:
    :param title:
    :param max_rows:
    :param title_color:
    :param highlight_color:
    :param nav_color:
    :return:
    """
    return MenuChoose.choose(items, checked, title, max_rows, title_color, highlight_color, nav_color)


class MenuChoose(ABC):

    @classmethod
    def choose(
            cls,
            items: List[Any],
            checked: bool = True,
            title: str = 'Please select one',
            max_rows: int = 15,
            title_color: VtColors = VtColors.ORANGE,
            highlight_color: VtColors = VtColors.BLUE,
            nav_color: VtColors = VtColors.YELLOW) -> List[Any]:

        done = None
        show_from = 0
        sel_index = -1
        re_render = 1
        sel_options = []
        length = len(items)
        signal.signal(signal.SIGINT, MenuUtils.exit_app)

        if length > 0:
            sel_index = 0
            init_value = 1 if checked else 0
            show_to = max_rows - 1

            diff_index = show_to - show_from
            length = len(items)

            # Initialize all options
            sel_options = [init_value for _ in range(length)]

            sysout(f"%ED2%%HOM%{title_color.placeholder()}{title}")
            vt_print(Vt100.set_auto_wrap(False))
            vt_print('%HOM%%CUD(1)%%ED0%')
            vt_print(Vt100.save_cursor())

            # Wait for user interaction
            while not done:

                # Menu Renderization {
                if re_render:
                    cls.__render__(items, sel_options, show_from, show_to, sel_index, highlight_color)
                    sysout(
                        f"{nav_color.placeholder()} [Enter] Accept  [\u2191\u2193] Navigate  [Space] Mark  [I] Invert  [Q] Quit  [1..{str(length)}] Goto: %EL0%", end='')
                    vt_print(Vt100.set_show_cursor(True))
                    re_render = None
                # } Menu Renderization

                # Navigation input {
                keypress = Keyboard.read_keystroke()
                if keypress in [Keyboard.VK_q, Keyboard.VK_Q, Keyboard.VK_ESC]:
                    done = True
                    sel_index = -1
                    sysout('\n%NC%')
                else:
                    if keypress.isdigit():  # An index was typed
                        typed_index = keypress.value
                        sysout(f"{keypress.value}", end='')
                        index_len = 1
                        while len(typed_index) < len(str(length)):
                            numpress = Keyboard.read_keystroke()
                            if not numpress:
                                break
                            elif not re.match(r'^[0-9]*$', numpress.value):
                                typed_index = None
                                break
                            typed_index = f"{typed_index}{numpress.value if numpress else ''}"
                            sysout(f"{numpress.value if numpress else ''}", end='')
                            index_len += 1
                        # Erase the index typed by the user
                        sysout(f"%CUB({index_len})%%EL0%", end='')
                        if 1 <= int(typed_index) <= length:
                            show_to = int(typed_index)
                            if show_to <= diff_index:
                                show_to = diff_index
                            show_from = show_to - diff_index
                            sel_index = int(typed_index) - 1
                            re_render = 1
                    elif keypress == Keyboard.VK_SPACE:  # Space -> Mark option
                        if sel_options[sel_index] == 0:
                            sel_options[sel_index] = 1
                        else:
                            sel_options[sel_index] = 0
                        re_render = 1
                        continue
                    elif keypress in [Keyboard.VK_i, Keyboard.VK_I]:  # I -> Invert options
                        sel_options = [(0 if op == 1 else 1) for op in sel_options]
                        re_render = 1
                        continue
                    elif keypress == Keyboard.VK_UP:  # Cursor up
                        if sel_index == show_from and show_from > 0:
                            show_from -= 1
                            show_to -= 1
                        elif sel_index == 0:
                            continue
                        if sel_index - 1 >= 0:
                            sel_index -= 1
                            re_render = 1
                    elif keypress == Keyboard.VK_DOWN:  # Cursor down
                        if sel_index + 1 == show_to and show_to < length:
                            show_from += 1
                            show_to += 1
                        elif sel_index + 1 >= length:
                            continue
                        if sel_index + 1 < length:
                            sel_index += 1
                            re_render = 1
                    elif keypress == Keyboard.VK_ENTER:  # Enter
                        sysout('\n%NC%')
                        break
                # } Navigation input

        vt_print('%HOM%%ED2%%MOD(0)%')

        return [op for idx, op in enumerate(items) if sel_options[idx] == 1] if sel_index >= 0 else None

    @classmethod
    def __render__(
            cls,
            items: List[Any],
            sel_options: List[int],
            show_from: int,
            show_to: int,
            sel_index: int,
            highlight_color: VtColors = VtColors.BLUE) -> None:

        length = len(items)
        rows, columns = screen_size()
        vt_print(Vt100.set_show_cursor(False))
        # Restore the cursor to the home position
        vt_print(Vt100.restore_cursor())
        sysout('%NC%')

        for idx in range(show_from, show_to):
            selector = ' '
            mark = ' '
            if idx >= length:
                break  # When the number of items is lower than the max_rows, skip the other lines
            option_line = str(items[idx])[0:int(columns)]
            # Erase current line before repaint
            vt_print('%EL2%\r')
            # Print the selector if the index is current
            if idx == sel_index:
                vt_print(highlight_color.code())
                selector = '>'
            # Print the marker if the option is checked
            if sel_options[idx] == 1:
                mark = 'x'
            fmt = " {:>" + str(len(str(length))) + "}  {:>2} [{}] {}"
            sysout(fmt.format(idx + 1, selector, mark, option_line))
            # Check if the text fits the screen and print it, otherwise print '...'
            if len(option_line) >= int(columns):
                vt_print("%CUB(4)%%EL0%...")
                sysout('%NC%')
        sysout('\n')



import signal
import time
from abc import ABC
from typing import Any, Optional, Callable

from hspylib.core.exception.input_aborted_error import InputAbortedError
from hspylib.core.tools.commons import sysout, syserr
from hspylib.core.tools.validator import Validator
from hspylib.ui.cli.vt100.vt_colors import VtColors


class MenuUtils(ABC):

    @staticmethod
    def exit_app(
            exit_code: int = signal.SIGHUP,
            frame=None,
            exit_msg: str = "Done.") -> None:

        sysout(str(frame) if frame else '', end='')
        sysout(f"%HOM%%ED2%%NC%\n{exit_msg}\n")
        exit(exit_code if exit_code else 0)

    @staticmethod
    def print_error(
            message: str,
            argument: Any = None,
            wait_interval: int = 2,
            color: VtColors = VtColors.RED) -> None:

        syserr(f"{color.placeholder()}### Error: {message} \"{argument or ''}\"%NC%")
        time.sleep(wait_interval)
        sysout('%CUU(2)%%ED0%', end='')

    @staticmethod
    def print_warning(
            message: str,
            argument: str = None,
            wait_interval: int = 2,
            color: VtColors = VtColors.YELLOW) -> None:

        sysout(f"{color.placeholder()}### Warn: {message} \"{argument or ''}\"%NC%")
        time.sleep(wait_interval)
        sysout('%CUU(2)%%ED0%', end='')

    @staticmethod
    def prompt(
            prompt_msg: str = '',
            validator: Callable = None,
            default_value: Any = None,
            any_key: bool = False,
            on_blank_abort: bool = True,
            color: VtColors = VtColors.GREEN,
            end: str = ': ') -> Optional[Any]:

        valid = False
        input_data = None

        while not valid:
            try:
                colorized = VtColors.colorize(
                    '{}{}{}{}{}'.format(
                        color.placeholder(),
                        prompt_msg,
                        '[{}]'.format(default_value) if default_value else '',
                        end, VtColors.NC.placeholder())
                )
                input_data = input(colorized)
                if not validator:
                    valid = True
                elif default_value:
                    return default_value
                elif any_key:
                    return None
                elif Validator.is_not_blank(input_data):
                    valid, msg = validator(input_data)
                    if not valid:
                        MenuUtils.print_error("{}: ".format(msg), input_data)
                else:
                    if not on_blank_abort:
                        MenuUtils.print_error("Input can't be empty: ", input_data)
                    else:
                        raise InputAbortedError()
            except EOFError as err:
                MenuUtils.print_error("Input failed: ", str(err))
                break

        return input_data

    @staticmethod
    def wait_enter(
            wait_msg: str = 'Press [Enter] to continue ...',
            color: VtColors = VtColors.YELLOW) -> None:
        MenuUtils.prompt(wait_msg, any_key=True, color=color, end='')

    @staticmethod
    def title(title_str: str, color: VtColors = VtColors.YELLOW) -> None:
        sysout(f"%ED2%%HOM%\n{color.placeholder()}{title_str}\n")



from _ctypes_test import func


class EventBus:
    __buses = {}
    __subscribers = {}
    __events = []

    @staticmethod
    def get(bus_name: str):
        if bus_name in EventBus.__buses:
            return EventBus.__buses[bus_name]
        else:
            bus_instance = EventBus(bus_name)
            EventBus.__buses[bus_name] = bus_instance
            return bus_instance

    @staticmethod
    def __create_or_get(bus_name: str, event_name: str):
        cache_key = '{}.{}'.format(bus_name, event_name)
        if cache_key in EventBus.__subscribers:
            return EventBus.__subscribers[cache_key]
        else:
            subscriber = {'callbacks': []}
            EventBus.__subscribers[cache_key] = subscriber
            return subscriber

    def __init__(self, name: str):
        self.name = name

    def subscribe(self, event_name: str, callback: func):
        subscriber = EventBus.__create_or_get(self.name, event_name)
        subscriber['callbacks'].append(callback)

    def emit(self, event_name: str, **kwargs):
        self.__events.append({'event': event_name, 'kwargs': kwargs})
        while len(self.__events) > 0:
            event = self.__events.pop()
            cache_key = '{}.{}'.format(self.name, event['event'])
            subscribers = self.__subscribers[cache_key] if cache_key in self.__subscribers else None
            if subscribers and len(subscribers['callbacks']) > 0:
                for callback in subscribers['callbacks']:
                    callback(event['kwargs'])



import logging as log
from typing import Optional

from requests.structures import CaseInsensitiveDict

from hspylib.core.enum.http_method import HttpMethod
from hspylib.modules.fetch.fetch import fetch
from hspylib.modules.fetch.http_response import HttpResponse


def soap_call(
        url: str,
        method: HttpMethod,
        data: str,
        headers: Optional[CaseInsensitiveDict]) -> Optional[HttpResponse]:
    all_headers = {} if not headers else headers
    all_headers.update({
        "Content-Type": "text/xml",
        "Accept": "*/*"
    })
    log.info('Processing SOAP {} {} -> {} \n{}'.format(all_headers, method, url, data if data else ''))
    response = fetch(url=url, method=method, headers=all_headers, body=data)
    log.info(
        'Response <=  Status: {}  Payload: {}'.format(
            response.status_code, response.body if response.body else '<None>'))

    return response



from requests.models import Response, CaseInsensitiveDict

from hspylib.core.enum.charset import Charset
from hspylib.core.enum.content_type import ContentType
from hspylib.core.enum.http_code import HttpCode
from hspylib.core.enum.http_method import HttpMethod


class HttpResponse:
    @staticmethod
    def of(response: Response):
        return HttpResponse(
            HttpMethod[response.request.method],
            response.url,
            HttpCode(response.status_code),
            response.text,
            response.headers,
            Charset(str(response.encoding).upper()) if response.encoding else None,
        )

    def __init__(self,
                 method: HttpMethod,
                 url: str,
                 status_code: HttpCode = None,
                 body: str = None,
                 headers: CaseInsensitiveDict = None,
                 encoding: Charset = Charset.UTF_8,
                 content_type=ContentType.APPLICATION_JSON):
        self.method = method
        self.url = url
        self.status_code = status_code
        self.body = body
        self.headers = headers
        self.encoding = encoding
        self.content_type = content_type
        if self.content_type:
            self.content_type.charset = self.encoding



#!/usr/bin/env python3
import getopt
import os
import pathlib
import subprocess
import sys
from typing import List

from hspylib.modules.fetch.fetch import get

from hspylib.core.enum.enumeration import Enumeration
from hspylib.core.enum.http_code import HttpCode
from hspylib.core.meta.singleton import Singleton
from hspylib.core.tools.commons import __version__, sysout, syserr

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The directory containing all template files
TEMPLATES = (HERE / "templates")

# The directory containing the welcome message
WELCOME = (HERE / "welcome.txt")

VERSION = __version__(f"{HERE}/.version")

USAGE = """
Usage: hspylib [options] <arguments>

    HSPyLib Manager v{} - Manage HSPyLib applications.

    Options:
      -v  |  --version                                      : Display current program version.
      -h  |     --help                                      : Display this help message.
      -c  |     --create <type> <app_name> [dest_dir]     : Manage an HSPyLib application.

    Arguments:
      type      : One of basic|gradle|git|all. "basic" is going to scaffold a basic app based on HSPyLib application 
                  structure. "gradle" is going to initialize you project with gradle (requires gradle installed). 
                  "git" is going to initialize a git repository. "all" is going to create a gradle project and also
                  initialize a git repository.

""".format('.'.join(map(str, VERSION)))

GRADLE_PROPS = f"""
project.ext.set("projectVersion", '0.1.0')
project.ext.set("pythonVersion", '3')
project.ext.set("pyrccVersion", '5')
"""


class Main(metaclass=Singleton):

    class AppType(Enumeration):
        BASIC = 1
        GRADLE = 2
        GIT = 4
        ALL = 8

    def __init__(self):
        super().__init__()
        self.app_dir = None
        self.init_gradle = False
        self.init_git = False

    def run(self, arguments: List[str]) -> None:
        if len(arguments) == 0:
            welcome = WELCOME.read_text()
            sysout(f"{welcome}")
            sysout(USAGE)
        else:
            try:
                opts, args = getopt.getopt(arguments, 'hvc:', ['--help', '--version', '--create'])
                for op, arg in opts:
                    if op in ['-h', '--help']:
                        sysout(USAGE)
                    elif op in ['-v', '--version']:
                        sysout('.'.join(map(str, VERSION)))
                    elif op in ['-c', '--create']:
                        assert arg.lower() in ['basic', 'gradle', 'git', 'all'], f'Invalid type: {arg}'
                        assert len(args) > 0, f'Invalid arguments: {str(args)}'
                        self._create_app(
                            Main.AppType.value_of(arg, ignore_case=True),
                            args[0],
                            args[1] if len(args) > 1 else os.environ.get('HOME'))
            except getopt.GetoptError as err:
                sysout(f"%RED%### Unhandled option: {str(err)}")
                sysout(USAGE)

    def _create_app(self, app_type: AppType, app_name: str, dest_dir: str):
        sysout(f'Creating app: {app_name} -> {dest_dir} ...')
        try:
            assert os.path.exists(dest_dir), f'Destination not found: {dest_dir}'
            self.app_dir = f'{dest_dir}/{app_name}'
            sysout(f'App: {app_name}')
            self._mkdir('')
            self._mkdir('src')
            self._mkdir('src/test')
            self._mkdir('src/main')
            self._mkfile('src/main/__main__.py', (TEMPLATES / "tpl-main.py").read_text())
            self._mkfile('src/main/.version', '0.1.0')
            self._mkdir('src/main/resources')
            self._mkfile('src/main/resources/application.properties', '# Main application property file')
            self._mkdir('src/main/resources/log')
            self._mkfile('README.md', f'# {app_name}')
            self._mkfile('.env', '# Type in here the environment variables your app requires')
            self._mkfile('run-it.sh', (TEMPLATES / "tpl-run-it.sh").read_text())
            os.chmod(f'{self.app_dir}/run-it.sh', 0o755)
            if app_type in [Main.AppType.GRADLE, Main.AppType.ALL]:
                self._init_gradle(app_name)
            if app_type in [Main.AppType.GIT, Main.AppType.ALL]:
                self._init_git()
        except OSError as err:
            syserr(f"Creation of the application {dest_dir}/{app_name} failed")
            syserr(str(err))
        else:
            sysout(f"Successfully created the application {dest_dir}/{app_name}")

    def _mkdir(self, dirname: str):
        dir_path = f"{self.app_dir}/{dirname}"
        sysout(f'  |- {dir_path}')
        os.mkdir(dir_path)

    def _mkfile(self, filename: str, contents: str):
        file_path = f"{self.app_dir}/{filename}"
        sysout(f'  |- {file_path}')
        with open(f'{file_path}', 'w') as fh:
            fh.write(contents)

    def _init_gradle(self, app_name: str):
        sysout('Initializing gradle project')
        args = ['gradle', 'init', '--project-name', app_name, '--type', 'basic', '--dsl', 'groovy']
        result = subprocess.run(args, capture_output=True, text=True, cwd=self.app_dir).stdout
        sysout('Gradle execution result: {}'.format(result))
        sysout('Downloading gradle extensions')
        self._download_ext('badges.gradle')
        self._download_ext('build-info.gradle')
        self._download_ext('docker.gradle')
        self._download_ext('oracle.gradle')
        self._download_ext('pypi-publish.gradle')
        self._download_ext('python.gradle')
        self._mkfile('properties.gradle', GRADLE_PROPS)
        self._mkfile(f'build.gradle', (TEMPLATES / "tpl-build.gradle").read_text())
        self._mkfile(f'gradle/dependencies.gradle', (TEMPLATES / "tpl-dependencies.gradle").read_text())

    def _download_ext(self, extension: str):
        resp = get(f'https://raw.githubusercontent.com/yorevs/hspylib/master/gradle/{extension}')
        assert resp.status_code == HttpCode.OK, f'Unable to download {extension}'
        self._mkfile(f'gradle/{extension}', resp.body)

    def _init_git(self):
        self._mkfile(f'src/main/resources/log/.gitkeep', '')
        self._mkfile(f'.gitignore', (TEMPLATES / "tpl.gitignore").read_text())
        sysout('Initializing git repository')
        result = subprocess.run(['git', 'init'], capture_output=True, text=True, cwd=self.app_dir).stdout
        sysout('Git init result: {}'.format(result))
        sysout('Creating first commit')
        subprocess.run(['git', 'add', '.'], capture_output=True, text=True, cwd=self.app_dir)
        result = subprocess.run(
            ['git', 'commit', '-m', 'First commit [@HSPyLib]'],
            capture_output=True, text=True, cwd=self.app_dir).stdout
        sysout('Git commit result: {}'.format(result))


# Application entry point
if __name__ == "__main__":
    """Application entry point"""
    Main().INSTANCE.run(sys.argv[1:])



import random
import re

from hspylib.core.tools.commons import get_or_default


def fit_text(text: str, width: int) -> str:
    return text if len(text) <= width else text[0:width-3] + '...'


def rand_string(choices: str, length: int) -> str:
    return ''.join(random.choices(choices, k=length))


class TextHelper:
    @staticmethod
    def justified_left(string: str, width: int, fill: str = ' ') -> str:
        return string.ljust(width, fill)

    @staticmethod
    def justified_center(string: str, width: int, fill: str = ' ') -> str:
        return string.center(width, fill)

    @staticmethod
    def justified_right(string: str, width: int, fill: str = ' ') -> str:
        return string.rjust(width, fill)

    @staticmethod
    def uppercase(string: str) -> str:
        return string.upper()

    @staticmethod
    def lowercase(string: str) -> str:
        return string.lower()

    @staticmethod
    def camelcase(string: str) -> str:
        return string.capitalize()

    @staticmethod
    def cut(string: str, index: int) -> str:
        result = tuple(re.split(r' +', string))
        return get_or_default(result, index)


class TextAlignment:
    """
    Table cell text justification helper.
    """
    LEFT = TextHelper.justified_left
    CENTER = TextHelper.justified_center
    RIGHT = TextHelper.justified_right


class TextCase:
    """
    Table cell text justification helper.
    """
    UPPER_CASE = TextHelper.uppercase
    LOWER_CASE = TextHelper.lowercase
    CAMEL_CASE = TextHelper.camelcase



from abc import ABC
from typing import Type, Optional

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QWidget, QToolButton, QStackedWidget, QLineEdit, QTableView, QComboBox, QSpinBox, \
    QDialogButtonBox, QDoubleSpinBox, QDateEdit, QLabel, QToolBox


class QtFinder(ABC):
    @staticmethod
    def find_widget(window: QWidget, widget: Type, name: str) -> Optional[QObject]:
        return window.findChild(widget, name)

    def __init__(self, window: QWidget):
        super().__init__()
        self.window = window

    def find_stacked_widget(self, name: str) -> Optional[QStackedWidget]:
        return QtFinder.find_widget(self.window, QStackedWidget, name)

    def find_tool_box(self, name: str) -> Optional[QToolBox]:
        return QtFinder.find_widget(self.window, QToolBox, name)

    def find_label(self, name: str) -> Optional[QLabel]:
        return QtFinder.find_widget(self.window, QLabel, name)

    def find_tool_button(self, name: str) -> Optional[QToolButton]:
        return QtFinder.find_widget(self.window, QToolButton, name)

    def find_table_view(self, name: str) -> Optional[QTableView]:
        return QtFinder.find_widget(self.window, QTableView, name)

    def find_line_edit(self, name: str) -> Optional[QLineEdit]:
        return QtFinder.find_widget(self.window, QLineEdit, name)

    def find_date_edit(self, name: str) -> Optional[QDateEdit]:
        return QtFinder.find_widget(self.window, QDateEdit, name)

    def find_combo_box(self, name: str) -> Optional[QComboBox]:
        return QtFinder.find_widget(self.window, QComboBox, name)

    def find_spin_box(self, name: str) -> Optional[QSpinBox]:
        return QtFinder.find_widget(self.window, QSpinBox, name)

    def find_double_spin_box(self, name: str) -> Optional[QDoubleSpinBox]:
        return QtFinder.find_widget(self.window, QDoubleSpinBox, name)

    def find_button_box(self, name: str) -> Optional[QDialogButtonBox]:
        return QtFinder.find_widget(self.window, QDialogButtonBox, name)



from typing import Any


class MenuOption:
    def __init__(self, parent: Any, option_index: int, option_text: str):
        self.parent = parent
        self.option_index = option_index
        self.option_text = option_text
        self.action_trigger = lambda s: print(f"Option: {self.option_index}-{self.option_text} selected!")

    def on_trigger(self, action_trigger) -> Any:
        self.action_trigger = action_trigger
        return self.parent

    def __str__(self) -> str:
        return "%GREEN%[{}]%NC% {}".format(self.option_index, self.option_text)



from requests.structures import CaseInsensitiveDict

from hspylib.core.enum.charset import Charset
from hspylib.core.enum.content_type import ContentType
from hspylib.core.enum.http_code import HttpCode
from hspylib.core.enum.http_method import HttpMethod
from hspylib.modules.fetch.http_response import HttpResponse


class MockResponse(HttpResponse):
    def __init__(self,
                 parent,
                 method: HttpMethod,
                 url: str,
                 status_code: HttpCode = None,
                 body: str = None,
                 headers=None,
                 encoding: Charset = Charset.UTF_8,
                 content_type=ContentType.APPLICATION_JSON):

        super().__init__(method, url, status_code, body, headers, encoding, content_type)
        self.parent = parent
        self.received_body = False

    def then_return(self,
                    code: HttpCode,
                    body: str = None,
                    headers=None,
                    encoding: Charset = Charset.UTF_8,
                    content_type=ContentType.APPLICATION_JSON):

        response = self.parent.mock(self.method, self.url)
        response.status_code = code
        response.body = body
        response.headers = headers if headers else []
        response.encoding = encoding
        response.content_type = content_type
        if response.content_type:
            response.content_type.charset = encoding
        return self.parent

    def then_return_with_received_body(self,
                                       code: HttpCode,
                                       headers: CaseInsensitiveDict = None,
                                       encoding: Charset = Charset.UTF_8,
                                       content_type=ContentType.APPLICATION_JSON):
        response = self.parent.mock(self.method, self.url)
        response.received_body = True
        response.body = None
        response.status_code = code
        response.headers = headers if headers else []
        response.encoding = encoding
        response.content_type = content_type
        if response.content_type:
            response.content_type.charset = encoding
        return self.parent



import getopt
import logging as log
import signal
import sys
from typing import List, Callable, Optional, Tuple

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.meta.singleton import Singleton
from hspylib.core.tools.commons import sysout
from hspylib.ui.cli.app.option import Option


class Application(metaclass=Singleton):
    """TODO"""

    def __init__(
            self,
            app_name: str,
            app_version: Tuple[int, int, int],
            app_usage: str = "Usage: __main__.py [options] <arguments>",
            source_dir: str = None,
            resource_dir: str = None,
            log_dir: str = None):

        signal.signal(signal.SIGINT, self.exit_handler)
        self.app_name = app_name
        self.app_version = app_version
        self.app_usage = app_usage
        self.options = []
        self.args = None
        self.configs = AppConfigs(
            source_root=source_dir if source_dir else '.',
            resource_dir=resource_dir,
            log_dir=log_dir
        )
        self.with_option('h', 'help', handler=self.usage)
        self.with_option('v', 'version', handler=self.version)

    def main(self, *args, **kwargs):
        """TODO"""
        pass

    def run(self, *args, **kwargs):
        """TODO"""
        self.main(*args, **kwargs)
        self.exit_handler()

    def cleanup(self):
        """TODO"""
        log.info('Cleanup handler called. nothing to cleanup')

    def exit_handler(self, signum=0, frame=None, clear_screen: bool = False) -> None:
        """
        Handle interruptions to shutdown gracefully
        :param signum: The signal number or the exit code
        :param frame: The frame raised by the signal
        :param clear_screen: Whether to clean the screen before execution or not
        """
        if frame is not None:
            log.warn('Signal handler hooked signum={} frame={}'.format(signum, frame))
            exit_code = 3
        else:
            log.info('Exit handler called')
            exit_code = signum
        self.cleanup()
        if clear_screen:
            sysout('%ED2%%HOM%')
        sys.exit(exit_code)

    def usage(self, *args) -> None:
        """Display the usage message and exit with the specified code ( or zero as default )
        :param args: The application arguments
        """
        sysout(self.app_usage)
        self.exit_handler(args[0] or 0)

    def version(self, *args) -> None:
        """Display the current program version and exit
        :param args: The application arguments
        """
        sysout('{} v{}'.format(self.app_name, '.'.join(map(str, self.app_version))))
        self.exit_handler()

    def parse_arguments(self, arguments: List[str]) -> None:
        """ Handle program arguments and options. Short opts: -<C>, Long opts: --<Word>
        :param arguments: The list of unparsed program arguments passed by the command line
        """
        try:
            if self.options:
                opts, self.args = getopt.getopt(arguments, self.__shortopts__(), self.__longopts__())
                assert self.__has_valid_options__(opts), f'Invalid number of options: {len(opts)}'
                for op, arg in opts:
                    opt = self.__getopt__(op)
                    if opt and opt.handler:
                        opt.handler(arg)
                    else:
                        assert False, f"Unhandled option: {op}"
        except getopt.GetoptError as err:
            sysout(f"%RED%### Unhandled option: {str(err)}")
            self.usage(1)
        except AssertionError as err:
            sysout(f"%RED%### {str(err)}")
            self.usage(1)

    def with_option(
            self,
            shortopt: chr,
            longopt: str,
            has_argument: bool = False,
            handler: Callable = None,
            required: bool = False):
        self.options.append(Option(shortopt, longopt, has_argument, handler, required))

    def __has_valid_options__(self, options: List[str]):
        if self.__reqopts__() > 0 and not len(options) and len(self.options):
            self.usage(1)
        elif len(options) < self.__reqopts__():
            return any(opt[0].strip() in ['-v', '--version', '-h', '--help'] for opt in options)
        return True

    def __shortopts__(self) -> str:
        str_all_opts = ''.join([lop.shortopt.replace('-', '') for lop in self.options])
        return str_all_opts

    def __longopts__(self) -> List[str]:
        str_all_opts = [lop.longopt.replace('-', '') + ' ' for lop in self.options]
        return str_all_opts

    def __getopt__(self, item: str) -> Optional[Option]:
        return next((op for op in self.options if op.is_eq(item)), None)

    def __reqopts__(self) -> int:
        return sum(opt.required for opt in self.options)



import sys
import uuid
import logging as log
from abc import abstractmethod
from typing import Optional, Tuple, List

import pymysql
from pymysql.err import OperationalError, ProgrammingError
from requests.structures import CaseInsensitiveDict as SqlFilter

from hspylib.core.crud.db.db_repository import DBRepository
from hspylib.core.crud.db.sql_factory import SqlFactory
from hspylib.core.model.entity import Entity


class MySqlRepository(DBRepository):
    _connections = {}

    def __init__(self):
        super().__init__()
        self._connector = None
        self._cursor = None
        self._sql_factory = SqlFactory()

    def is_connected(self):
        return self._connector is not None

    def connect(self):
        if not self.is_connected():
            cache_key = self.__str__()
            if cache_key in MySqlRepository._connections:
                self._connector = MySqlRepository._connections[cache_key]
                self._cursor = self._connector.cursor()
                assert self.is_connected(), "Not connected to the database"
            else:
                try:
                    self._connector = pymysql.connect(
                        host=self.hostname,
                        user=self.user,
                        port=self.port,
                        password=self.password,
                        database=self.database
                    )
                    assert self.is_connected(), "Unable to connect to the database"
                    self._cursor = self._connector.cursor()
                    log.debug('Connection to {} established'.format(str(self)))
                    MySqlRepository._connections[cache_key] = self._connector
                except (OperationalError, ConnectionRefusedError):
                    log.error('Unable to connect to {}'.format(str(self)))
                    sys.exit(1)

    def disconnect(self):
        if self.is_connected():
            cache_key = self.__str__()
            self._connector.close()
            self._connector = None
            del MySqlRepository._connections[cache_key]
            log.debug('Disconnected from {}.'.format(str(self)))
        else:
            log.error('Unable to disconnect from {}'.format(str(self)))
            sys.exit(1)

        return self._connector

    def insert(self, entity: Entity):
        if self.is_connected():
            entity.uuid = entity.uuid if entity.uuid is not None else str(uuid.uuid4())
            stm = self._sql_factory\
                .insert(entity)\
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: {}'.format(stm))
            self._cursor.execute(stm)
        else:
            log.error('Not connected to database.')

    def update(self, entity: Entity):
        if self.is_connected():
            stm = self._sql_factory\
                .update(entity, filters=SqlFilter({"UUID": '{}'.format(entity.uuid)}))\
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: {}'.format(stm))
            self._cursor.execute(stm)
        else:
            log.error('Not connected to database.')

    def delete(self, entity: Entity):
        if self.is_connected():
            stm = self._sql_factory\
                .delete(filters=SqlFilter({"UUID": '{}'.format(entity.uuid)}))\
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: {}'.format(stm))
            self._cursor.execute(stm)

    def find_all(self, column_set: List[str] = None, sql_filters: SqlFilter = None) -> Optional[list]:
        if self.is_connected():
            stm = self._sql_factory\
                .select(column_set=column_set, filters=sql_filters)\
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: {}'.format(stm))
            self._cursor.execute(stm)
            result = self._cursor.fetchall()
            return list(map(self.row_to_entity, result)) if result else None
        else:
            log.error('Not connected to database.')

    def find_by_id(self, column_set: List[str] = None, entity_id: str = None) -> Optional[Entity]:
        if self.is_connected():
            if entity_id:
                stm = self._sql_factory\
                    .select(column_set=column_set, filters=SqlFilter({"UUID": '{}'.format(entity_id)}))\
                    .replace(':tableName', self.table_name())
                log.debug('Executing SQL statement: {}'.format(stm))
                self._cursor.execute(stm)
                result = self._cursor.fetchall()
                if len(result) > 1:
                    raise ProgrammingError('Multiple results found')
                return self.row_to_entity(result[0]) if len(result) > 0 else None
            else:
                return None
        else:
            log.error('Not connected to database.')

    def execute(self, sql_statement: str, auto_commit: bool = True, *params):
        if self.is_connected():
            self._cursor.execute(sql_statement, params)
            log.debug('Executing SQL statement: {}'.format(sql_statement))
            if auto_commit:
                self.commit()
        else:
            log.error('Not connected to database.')

    def commit(self):
        log.debug('Committing database changes')
        self._connector.commit()

    def rollback(self):
        log.debug('Rolling back database changes')
        self._connector.rollback()

    @abstractmethod
    def row_to_entity(self, row: Tuple) -> Entity:
        pass

    @abstractmethod
    def table_name(self) -> str:
        pass



import signal
from typing import Optional

from hspylib.ui.cli.menu.menu import Menu
from hspylib.ui.cli.menu.menu_utils import MenuUtils


class MenuUi:
    def __init__(self, root: Optional[Menu]):
        self.done = False
        self.previous = None
        self.current = root
        self.next = None
        signal.signal(signal.SIGINT, MenuUtils.exit_app)

    def run(self) -> None:
        while not self.done:
            if self.current:
                self.next = self.current.execute()
                if self.next is None:
                    self.done = True
                else:
                    self.change_menu(self.next)
            else:
                self.done = True

    def change_menu(self, menu: Menu) -> None:
        self.previous = self.current
        self.current = menu



import logging as log
import os
import re
import sys
import termios
import tty
from typing import Type, List, Tuple, Any, Optional

from hspylib.core.tools.validator import Validator
from hspylib.ui.cli.vt100.vt_codes import VtCodes
from hspylib.ui.cli.vt100.vt_colors import VtColors

LOG_FMT = '{} {} {} {}{} {} '.format(
    '%(asctime)s',
    '[%(threadName)-10.10s]',
    '%(levelname)-5.5s',
    '%(filename)s::',
    '%(funcName)s(@Line:%(lineno)d)',
    '%(message)s'
)


def log_init(
        log_file: str,
        create_new: bool = True,
        f_mode: str = 'a',
        level: int = log.DEBUG,
        log_fmt: str = LOG_FMT):
    """Initialize the system logger
    :param log_file: TODO
    :param create_new:  TODO
    :param f_mode:  TODO
    :param level:  TODO
    :param log_fmt:  TODO
    """
    with open(log_file, 'w' if create_new else 'a'):
        os.utime(log_file, None)

    log.basicConfig(
        filename=log_file,
        format=log_fmt,
        level=level,
        filemode=f_mode)

    return log


def __version__(version_filepath: str = ".version") -> Optional[Tuple]:
    """Retrieve the version from the version file in the form: Tuple[major,minor,build]"""
    try:
        with open(version_filepath) as fh_version:
            return tuple(map(str.strip, fh_version.read().split('.')))
    except FileNotFoundError:
        return None


def __curdir__(filepath: str) -> str:
    """Retrieve the application directory"""
    return os.path.dirname(os.path.realpath(filepath))


def environ_name(property_name: str) -> str:
    """Retrieve the environment name of the specified property name"""
    return re.sub('[ -.]', '_', property_name).upper()


def sysout(string: str, end: str = '\n') -> None:
    """Print the unicode input_string decoding vt100 placeholders"""
    if Validator.is_not_blank(string):
        msg = VtColors.colorize(VtCodes.decode(f"{string}"))
        print(msg, file=sys.stdout, flush=True, end=end)


def syserr(string: str, end: str = '\n') -> None:
    """Print the unicode input_string decoding vt100 placeholders"""
    if Validator.is_not_blank(string):
        msg = VtColors.colorize(VtCodes.decode(f"{string}"))
        print(msg, file=sys.stderr, flush=True, end=end)


def class_attribute_names(clazz: Type) -> tuple:
    """TODO
    :param clazz:
    """
    return tuple(vars(clazz()).keys()) if clazz else None


def class_attribute_values(instance: dict) -> tuple:
    """TODO
    :param instance:
    """
    return tuple(instance.values()) if object else None


def split_and_filter(input_str: str, regex_filter: str = '.*', delimiter: str = '\n') -> List[str]:
    """TODO
    :param input_str:
    :param regex_filter:
    :param delimiter:
    """
    regex = re.compile(regex_filter)
    result_list = list(filter(regex.search, input_str.split(delimiter)))

    return result_list


def get_or_default(options: tuple, index: int, default_value=None) -> Optional[Any]:
    """Retrieve an item from the options list or None if index is out of range
    :param options: The available list of options
    :param index: The index of the item
    :param default_value: The default value if the index is out of range
    """
    return options[index] if index < len(options) else default_value


def get_by_key_or_default(options: dict, key: str, default_value=None) -> Optional[Any]:
    """Retrieve an item from the options list or None if key was not found
    :param options: The available list of options
    :param key: The key of the item
    :param default_value: The default value if the index is not found
    """
    return options[key] if key in options else default_value


def str_to_bool(string: str, true_values: List[str] = None) -> bool:
    """Convert a string to boolean
    :param string: The string to be converted
    :param true_values: The list of strings that will become True value
    """
    if true_values is None:
        true_values = [
            'true', 'on', 'yes', '1', 'y'
        ]
    return string.lower() in true_values if string else False


def safe_del_file(filename: str, on_not_found_except: bool = False) -> bool:
    """TODO
    :param filename:
    :param on_not_found_except:
    """
    if os.path.exists(filename):
        os.remove(filename)
        return True
    else:
        if on_not_found_except:
            raise FileNotFoundError('File was not found on the system: {}'.format(filename))
        else:
            return False


def file_is_not_empty(filename: str) -> bool:
    """TODO
    :param filename:
    """
    return os.path.exists(filename) and os.stat(filename).st_size > 0


def touch_file(filename: str) -> None:
    """TODO
    :param filename:
    """
    with open(filename, 'a'):
        os.utime(filename, None)


def screen_size() -> Optional[List[str]]:
    """TODO
    """

    if sys.stdout.isatty():
        return os.popen('stty size').read().split()
    else:
        return None


def set_enable_echo(enable: bool = True) -> None:
    """
    TODO
    :param enable:
    :return:
    """
    if sys.stdout.isatty():
        os.popen(f"stty {'echo -raw' if enable else 'raw -echo min 0'}").read()


def get_cursor_position() -> Optional[Tuple[int, int]]:
    """ Get the terminal cursor position
    Solution taken from:
    - https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3
    :return:
    """
    if sys.stdout.isatty():
        buf = ""
        stdin = sys.stdin.fileno()
        attrs = termios.tcgetattr(stdin)

        try:
            tty.setcbreak(stdin, termios.TCSANOW)
            sys.stdout.write("\x1b[6n")
            sys.stdout.flush()
            while True:
                buf += sys.stdin.read(1)
                if buf[-1] == "R":
                    break
        finally:
            termios.tcsetattr(stdin, termios.TCSANOW, attrs)
        try:
            matches = re.match(r"^\x1b\[(\d*);(\d*)R", buf)
            groups = matches.groups()
        except AttributeError:
            return None

        return int(groups[0]), int(groups[1])
    else:
        return None



class Singleton(type):
    __instances = {}

    def __call__(cls, *args, **kwargs):
        if not Singleton.has_instance(cls):
            instance = super(Singleton, cls).__call__(*args, **kwargs)
            cls.INSTANCE = instance
            cls.__instances[cls] = instance

        return cls.__instances[cls]

    @classmethod
    def instances(mcs):
        print(Singleton.__instances)

    @classmethod
    def has_instance(mcs, cls):
        return cls in cls.__instances



from hspylib.core.enum.enumeration import Enumeration


class HttpMethod(Enumeration):
    OPTIONS = 'menu_options'
    HEAD = 'head'
    GET = 'get'
    POST = 'post'
    PUT = 'put'
    PATCH = 'patch'
    DELETE = 'delete'



import struct

from idna.core import unicode

from hspylib.core.enum.enumeration import Enumeration


def awe_print(awesome_str: str) -> None:
    print(awesome_str + ' ', end='')


class Awesome(Enumeration):
    """
        Font awesome codes
        Full list of font awesome icons can be found here:
          - https://fontawesome.com/cheatsheet?from=io
    """

    @staticmethod
    def demo_unicodes():
        i = 0
        st_base = ['F{:03X}'.format(x) for x in range(0, 4095)]
        for n in st_base:
            hexa = unicode(struct.pack("!I", int(n, 16)), 'utf_32_be')
            endz = '\n' if i != 0 and i % 10 == 0 else ' '
            print('{} {}'.format(hexa, n), end=endz)
            i += 1

    @classmethod
    def demo_icons(cls):
        list(map(lambda e: awe_print(e), cls.values()))

    def __str__(self) -> str:
        return str(self.value)

    def placeholder(self) -> str:
        return f":{self.name}:"


if __name__ == '__main__':
    Awesome.demo_unicodes()



import os
import sys
from typing import List

from hspylib.core.tools.commons import __version__, __curdir__, sysout
from hspylib.ui.cli.app.application import Application


class Main(Application):
    """TODO"""

    # The application name
    APP_NAME = os.path.basename(__file__)

    # Usage message
    USAGE = f"Usage: {APP_NAME} <option> [arguments]"

    def __init__(self, app_name: str):
        super().__init__(app_name, __version__(), self.USAGE, __curdir__(__file__))

    def main(self, arguments: List[str]) -> None:
        """Run the application with the command line arguments"""
        sysout(f'Hello {self.APP_NAME}')


if __name__ == "__main__":
    """Application entry point"""
    Main('Application name').INSTANCE.run(sys.argv[1:])



from hspylib.ui.cli.icons.font_awesome.awesome import Awesome


class FormIcons(Awesome):
    """
        Form UI icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    ARROW_LEFT = '\uF060'       # 
    ARROW_RIGHT = '\uF061'      # 
    ARROW_UP = '\uF062'         # 
    ARROW_DOWN = '\uF063'       # 

    CHECK = '\uF00C'            # 
    ERROR = '\uf057'            # 
    CHECK_CIRCLE = '\uF058'     # 
    UNCHECK_CIRCLE = '\uF111'   # 
    CHECK_SQUARE = '\uF14A'     # 
    UNCHECK_SQUARE = '\uF0C8'   # 
    ON = '\uF205'               # 
    OFF = '\uF204'              # 

    FOLDER = '\uF74A'           # 
    FOLDER_OPEN = '\uF07C'      # 
    HIDDEN = '\uF070'           # 
    VISIBLE = '\uF06E'          # 
    LOCKED = '\uF023'           # 
    UNLOCKED = '\uF09C'         # 
    EDITABLE = '\uF044'         # 
    DELETE = '\uF014'           # 
    EXIT = '\uF08B'             # 


if __name__ == '__main__':
    FormIcons.demo_icons()



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QLabel


class ButtonLabel(QLabel):

    clicked = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent=parent)

    def mousePressEvent(self, ev):
        self.clicked.emit()



import ast
import logging as log
import os


class FileStorage:
    def __init__(self, filename: str):
        self.filename = filename
        self.data = []
        self.load()
        log.debug('File storage filename={} created and loaded entries={}'.format(filename, len(self.data)))

    def load(self):
        mode = 'r+' if os.path.exists(self.filename) else 'w+'
        with open(self.filename, mode) as f_local_db:
            lines = f_local_db.read()
            if lines:
                saved_data = ast.literal_eval(lines)
                self.data = saved_data

    def commit(self):
        with open(self.filename, 'w') as f_local_db:
            f_local_db.write(str(self.data))
        log.debug('File storage filename={} committed entries={}'.format(self.filename, len(self.data)))

    def truncate(self):
        open(self.filename, 'w').close()
        log.warn('File storage filename={} was truncated'.format(self.filename))



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame


class Panel(QFrame):

    keyPressed = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent=parent)

    def keyPressEvent(self, event):
        self.keyPressed.emit(event.key())



from hspylib.core.meta.singleton import Singleton
from hspylib.ui.cli.factory.menu_entry import MenuEntry
from hspylib.ui.cli.factory.menu_option import MenuOption
from hspylib.ui.cli.menu.menu import Menu


class MenuFactory(metaclass=Singleton):

    class MenuBuilder:
        def __init__(self, parent_menu: Menu = None, menu_title: str = None):
            self.parent_menu = parent_menu
            self.menu_title = menu_title
            self.menu_options = {}

        def with_option(self, option_text: str) -> MenuOption:
            option_index = len(self.menu_options)
            option = MenuOption(self, option_index, option_text)
            self.menu_options[str(option_index)] = option
            return option

        def build(self) -> Menu:
            return MenuEntry(self.parent_menu, self.menu_options, self.menu_title)

    @staticmethod
    def create(parent_menu: Menu = None, menu_title: str = None) -> MenuBuilder:
        return MenuFactory.MenuBuilder(parent_menu, menu_title)



from abc import abstractmethod
from typing import Tuple

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.crud.crud_repository import CrudRepository
from hspylib.core.model.entity import Entity


class DBRepository(CrudRepository):
    def __init__(self):
        super().__init__()
        self.hostname = AppConfigs.INSTANCE['datasource.hostname']
        self.port = AppConfigs.INSTANCE.get_int('datasource.port')
        self.user = AppConfigs.INSTANCE['datasource.username']
        self.password = AppConfigs.INSTANCE['datasource.password']
        self.database = AppConfigs.INSTANCE['datasource.database']
        self.logger = AppConfigs.INSTANCE.logger()

    def __str__(self):
        return "{}@{}:{}/{}".format(self.user, self.hostname, self.port, self.database)

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def is_connected(self):
        pass

    @abstractmethod
    def execute(self, sql_statement: str, auto_commit: bool = True, *params):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def rollback(self):
        pass

    @abstractmethod
    def row_to_entity(self, row: Tuple) -> Entity:
        pass

    @abstractmethod
    def table_name(self) -> str:
        pass



from emoji.core import emojize

from hspylib.core.enum.enumeration import Enumeration


def emj_print(emoji_str: str) -> None:
    print(emojize(emoji_str) + ' ', end='')


class Emoji(Enumeration):
    """
        Emoji codes
        Full list of emojis can be found here:
          - https://unicode.org/emoji/charts/emoji-list.html
    """
    def __str__(self) -> str:
        return str(self.value)

    def placeholder(self) -> str:
        return f":{self.name}:"



from abc import ABC, abstractmethod

from PyQt5.QtWidgets import QWidget

from hspylib.ui.qt.qt_finder import QtFinder


class QtView(ABC):
    def __init__(self, window: QWidget, parent=None):
        super().__init__()
        self.window = window
        self.parent = parent
        self.qt = QtFinder(self.window)

    @abstractmethod
    def setup_ui(self):
        pass



from http.server import HTTPServer
from random import randint
from threading import Thread
from typing import Tuple, Optional

from hspylib.core.enum.http_method import HttpMethod
from hspylib.modules.mock.mock_request import MockResponse
from hspylib.modules.mock.mock_server_handler import MockServerHandler


class MockServer(HTTPServer):

    RANDOM_PORT = randint(49152, 65535)

    def mock(self, method: HttpMethod, url: str) -> Optional[MockResponse]:
        try:
            return self.__mocks[method][url]
        except KeyError:
            return None

    def __init__(self, hostname: str, port: int):
        self.__mocks = {}
        self.hostname = hostname
        self.port = port
        self.version = '0.9.0'
        super().__init__(self.address(), MockServerHandler)

    def address(self) -> Tuple[str, int]:
        return self.hostname, self.port

    def is_allowed(self, method: HttpMethod):
        return method in self.__mocks

    def start(self):
        runner = ServerThread(self)
        runner.start()

    def stop(self):
        self.shutdown()
        self.server_close()

    def when_request(self, method: HttpMethod, url: str):
        request = self.mock(method=method, url=url) or MockResponse(self, method, url)
        self.__mocks[method] = self.__mocks[method] \
            if method in self.__mocks else {}
        self.__mocks[method][url] = request
        return request


class ServerThread(Thread):
    def __init__(self, parent: MockServer):
        super().__init__()
        self.parent = parent

    def run(self) -> None:
        self.parent.serve_forever()



import logging as log
import re
import uuid
from abc import abstractmethod
from typing import Optional, List

from hspylib.core.crud.crud_repository import CrudRepository
from hspylib.core.crud.file.file_storage import FileStorage
from hspylib.core.model.entity import Entity


class FileRepository(CrudRepository):
    __storages = {}

    @staticmethod
    def check_criteria(partial_value, whole_value):
        """TODO
        :param partial_value:
        :param whole_value:
        :return:
        """
        if isinstance(whole_value, str):
            return str(partial_value).upper() in whole_value.upper()
        elif isinstance(whole_value, int):
            return int(partial_value) == whole_value
        elif isinstance(whole_value, float):
            return float(partial_value) == whole_value
        elif isinstance(whole_value, bool):
            return bool(partial_value) == whole_value
        else:
            return False

    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.storage = self.__create_or_get()

    def __str__(self):
        return str(self.storage.data)

    def insert(self, entity: Entity):
        """TODO
        :param entity:
        """
        entity.uuid = entity.uuid if entity.uuid else str(uuid.uuid4())
        self.storage.data.append(entity.to_dict())
        self.storage.commit()
        log.debug("{} has been inserted !".format(entity.__class__.__name__))

    def update(self, entity: Entity):
        """TODO
        :param entity:
        """
        for index, next_entry in enumerate(self.storage.data):
            if next_entry['uuid'] == entity.uuid:
                self.storage.data[index] = entity.to_dict()
                self.storage.commit()
                log.debug("{} has been updated !".format(entity.__class__.__name__))

    def delete(self, entity: Entity):
        """TODO
        :param entity:
        """
        for index, next_entry in enumerate(self.storage.data):
            if next_entry['uuid'] == entity.uuid:
                self.storage.data.remove(self.storage.data[index])
                self.storage.commit()
                log.debug("{} has been deleted !".format(entity.__class__.__name__))

    def find_all(self, filters: str = None) -> List[Entity]:
        """TODO
        :param filters:
        """
        self.storage.load()
        if filters:
            file_filters = filters.split(',')
            filtered = []
            for next_filter in file_filters:
                fields = re.split('=|>|<|>=|<=|==|!=', next_filter)
                try:
                    found = [
                        self.dict_to_entity(data) for data in self.storage.data if
                        self.check_criteria(fields[1], data[fields[0]])
                    ]
                except KeyError:
                    continue
                except IndexError:
                    continue
                filtered.extend(found)
            return filtered
        else:
            return [self.dict_to_entity(data) for data in self.storage.data]

    def find_by_id(self, entity_id: uuid.UUID) -> Optional[Entity]:
        """TODO
        :param entity_id:
        """
        self.storage.load()
        if entity_id:
            result = [data for data in self.storage.data if entity_id == data['uuid']]
            assert len(result) <= 1, "Multiple results found with entity_id={}".format(entity_id)

            return self.dict_to_entity(result[0]) if len(result) > 0 else None
        else:
            return None

    @abstractmethod
    def dict_to_entity(self, row: dict) -> Entity:
        """TODO
        :param row:
        """
        pass

    def __create_or_get(self):
        """TODO"""
        if self.filename in FileRepository.__storages:
            return FileRepository.__storages[self.filename]
        else:
            FileRepository.__storages[self.filename] = FileStorage(self.filename)
            return FileRepository.__storages[self.filename]



#!/usr/bin/env python3
import re
import signal
import time
from typing import Any, List

from hspylib.core.tools.commons import sysout, set_enable_echo, get_cursor_position
from hspylib.core.tools.keyboard import Keyboard
from hspylib.ui.cli.icons.font_awesome.ui_compose.form_icons import FormIcons
from hspylib.ui.cli.menu.menu_utils import MenuUtils
from hspylib.ui.cli.vt100.vt_100 import Vt100
from hspylib.ui.cli.vt100.vt_codes import vt_print
from hspylib.ui.cli.vt100.vt_colors import VtColors


def minput(
        form_fields: List[Any],
        title: str = 'Please fill all fields of the form fields below',
        title_color: VtColors = VtColors.ORANGE,
        nav_color: VtColors = VtColors.YELLOW) -> Any:
    """
    TODO
    :param form_fields:
    :param title:
    :param title_color:
    :param nav_color:
    :return:
    """
    return MenuInput(form_fields).input(title, title_color, nav_color)


class MenuInput:

    __mi_modes__ = ['input', 'password', 'checkbox']
    __mi_kinds__ = ['letter', 'number', 'word', 'any']
    __mi_access_types__ = ['read-only', 'read-write']

    SELECTED_BG = '%MOD(44)%'

    @staticmethod
    class Field:
        def __init__(
                self,
                label: str = None,
                mode: str = 'input',
                kind: str = 'any',
                min_length: int = 0,
                max_length: int = 30,
                access_type: str = 'read-write',
                value: Any = None):

            self.label = label
            self.mode = mode
            self.kind = kind
            self.min_length = min_length
            self.max_length = max_length
            self.access_type = access_type
            self.value = value

        def __str__(self) -> str:
            return str(self.__dict__)

        def can_write(self) -> bool:
            return 'read-write' == self.access_type

        def val_regex(self, min_length: int, max_length: int) -> str:
            if 'letter' == self.kind:
                regex = r'^[a-zA-Z]{' + str(min_length) + ',' + str(max_length) + '}$'
            elif 'number' == self.kind:
                regex = r'^[0-9]{' + str(min_length) + ',' + str(max_length) + '}$'
            elif 'word' == self.kind:
                regex = r'^[a-zA-Z0-9 _]{' + str(min_length) + ',' + str(max_length) + '}$'
            else:
                regex = r'.{' + str(min_length) + ',' + str(max_length) + '}$'

            return regex

    @staticmethod
    class FormBuilder:
        def __init__(self):
            self.fields = []

        def field(self) -> Any:
            return MenuInput.FieldBuilder(self)

        def build(self) -> list:
            return self.fields

    @staticmethod
    class FieldBuilder:
        def __init__(self, parent: Any):
            self.parent = parent
            self.field = MenuInput.Field()

        def label(self, label: str):
            self.field.label = label
            return self

        def mode(self, mode: str):
            assert mode in MenuInput.__mi_modes__, \
                f"Not a valid mode: {mode}. Valid modes are: {str(MenuInput.__mi_modes__)}"
            self.field.mode = mode
            return self

        def kind(self, kind: str):
            assert kind in MenuInput.__mi_kinds__, \
                f"Not a valid kind: {kind}. Valid kinds are: {str(MenuInput.__mi_kinds__)}"
            self.field.kind = kind

            return self

        def min_max_length(self, min_length: int, max_length: int):
            assert max_length >= min_length, f"Not a valid field length: ({min_length}-{max_length})"
            assert max_length > 0 and min_length > 0, f"Not a valid field length: ({min_length}-{max_length})"
            self.field.min_length = min_length
            self.field.max_length = max_length
            return self

        def access_type(self, access_type: str):
            assert access_type in MenuInput.__mi_access_types__, \
                f"Not a valid access type {access_type}. Valid access types are: {str(MenuInput.__mi_access_types__)}"
            self.field.access_type = access_type
            return self

        def value(self, value: Any):
            re_valid = self.field.val_regex(0, len(str(value)))
            if value:
                assert re.match(re_valid, value), \
                    f"Not a valid value: \"{value}\". Valid regex is \"{re_valid}\""
            self.field.value = value
            return self

        def build(self):
            if self.field.mode == "checkbox":
                self.field.value = self.field.value if self.field.value in ['0', '1'] else 0
                self.field.min_length = self.field.max_length = 1
            self.field.label = self.field.label or 'Field'
            self.field.access_type = self.field.access_type or 'read-write'
            self.field.min_length = self.field.min_length or 1
            self.field.max_length = self.field.max_length or 30
            self.field.kind = self.field.kind or 'any'
            self.field.kind = self.field.kind or 'input'
            self.field.value = self.field.value or ''
            self.parent.fields.append(self.field)
            return self.parent

    @staticmethod
    def __detail_len__(field: Any) -> int:
        max_len = len(str(field.max_length))
        return 1 + (2 * max_len)

    @classmethod
    def mi_print(cls, size: int, text: str, prepend: str = None, end: str = ''):
        fmt = ('{}' if prepend else '') + "{:<" + str(size) + "} : "
        if prepend:
            vt_print(fmt.format(prepend, text), end=end)
        else:
            vt_print(fmt.format(text), end=end)

    @classmethod
    def builder(cls):
        return cls.FormBuilder()

    def __init__(self, all_fields: List[Any]):
        self.all_fields = all_fields
        self.all_pos = [(0, 0) for _ in all_fields]
        self.cur_field = None
        self.cur_row = 0
        self.cur_col = 0
        self.tab_index = 0
        self.err_msg = ''
        self.max_label_length = max([len(field.label) for field in all_fields])
        self.max_value_length = max([field.max_length for field in all_fields])
        self.max_detail_length = max([self.__detail_len__(field) for field in all_fields])
        self.done = None
        self.re_render = True

    def input(
            self,
            title: str = 'Please fill all fields of the form fields below',
            title_color: VtColors = VtColors.ORANGE,
            nav_color: VtColors = VtColors.YELLOW) -> List[Any]:

        ret_val = None
        length = len(self.all_fields)
        signal.signal(signal.SIGINT, MenuUtils.exit_app)
        signal.signal(signal.SIGHUP, MenuUtils.exit_app)

        if length > 0:
            sysout(f"%ED2%%HOM%{title_color.placeholder()}{title}")
            vt_print(Vt100.set_auto_wrap(False))
            vt_print('%HOM%%CUD(1)%%ED0%')
            vt_print(Vt100.save_cursor())

            # Wait for user interaction
            while not self.done and ret_val != Keyboard.VK_ENTER:
                # Menu Renderization {
                if self.re_render:
                    self.__render__(nav_color)
                # } Menu Renderization

                vt_print(Vt100.set_show_cursor(True))

                # Position the cursor to edit the current field
                if "checkbox" != self.cur_field.mode:
                    vt_print(f"%CUP({self.cur_row};{self.cur_col})%")
                else:
                    vt_print(f"%CUP({self.cur_row};{self.max_label_length + 6})%")

                # Navigation input {
                ret_val = self.__nav_input__()
                self.re_render = True
                # } Navigation input

        vt_print('%HOM%%ED2%%MOD(0)%')

        return self.all_fields if ret_val == Keyboard.VK_ENTER else []

    def __render__(self, nav_color: VtColors):
        icon = ''
        vt_print(Vt100.set_show_cursor(False))
        # Restore the cursor to the home position
        vt_print(Vt100.restore_cursor())
        sysout('%NC%')
        set_enable_echo()
        vt_print(Vt100.set_show_cursor(False))

        for idx, field in enumerate(self.all_fields):
            field_size = len(str(field.value))
            if self.tab_index != idx:
                MenuInput.mi_print(self.max_label_length, field.label)
            else:
                MenuInput.mi_print(self.max_label_length, field.label, MenuInput.SELECTED_BG)
                # Buffering the all positions to avoid calling get_cursor_pos
                f_pos = get_cursor_position() if self.all_pos[idx] == (0, 0) else self.all_pos[idx]
                if f_pos:
                    self.cur_row = f_pos[0]
                    self.cur_col = f_pos[1] + field_size
                    self.all_pos[idx] = f_pos
                # Keep the selected field on hand
                self.cur_field = field

            # Choose the icon to display
            if field.mode == "input":
                icon = FormIcons.EDITABLE
                MenuInput.mi_print(self.max_value_length, field.value)
            elif field.mode == "password":
                icon = FormIcons.HIDDEN
                MenuInput.mi_print(self.max_value_length, '*' * field_size)
            elif field.mode == "checkbox":
                icon = FormIcons.EDITABLE
                if field.value:
                    MenuInput.mi_print(self.max_value_length - 1, ' ', str(FormIcons.CHECK_SQUARE))
                else:
                    MenuInput.mi_print(self.max_value_length - 1, ' ', str(FormIcons.UNCHECK_SQUARE))
            if field.access_type == 'read-only':
                icon = FormIcons.LOCKED

            # Remaining/max characters
            padding = 1 - len(str(self.max_detail_length / 2))
            fmt = "{:<3}{:>" + str(padding) + "}/{:<" + str(padding) + "} %MOD(0)%"
            sysout(fmt.format(icon, field_size, field.max_length))

            # Display any previously set error message
            if self.tab_index == idx and self.err_msg:
                self.__display_error__()
                return

        sysout('\n')
        sysout(
            f"{nav_color.placeholder()}[Enter] Submit  [\u2191\u2193] Navigate  [Tab] Next  [Esc] Quit %EL0%", end='')
        self.re_render = False

    def __display_error__(self):
        err_offset = 12 + self.max_detail_length
        set_enable_echo(False)
        err_pos = self.max_label_length + self.max_value_length + err_offset
        vt_print(f"%CUP({self.cur_row};{err_pos})%")
        sysout(f"%RED% {FormIcons.ERROR}  {self.err_msg}", end='')
        dismiss_timeout = 1 + (len(self.err_msg) / 25)
        time.sleep(dismiss_timeout)
        # Remove the message after the timeout
        set_enable_echo()
        vt_print(f"%CUP({self.cur_row};{err_pos})%%EL0%")
        self.err_msg = ''

    def __nav_input__(self):
        length = len(self.all_fields)
        keypress = Keyboard.read_keystroke()

        if not keypress:
            return None
        elif keypress == Keyboard.VK_ESC:
            self.done = True
            sysout('\n%NC%')
        else:
            if keypress == Keyboard.VK_TAB:  # Handle TAB
                # Validate and move next. First case statement because next one also captures it
                min_len = self.cur_field.min_length
                if min_len <= len(str(self.cur_field.value)):
                    if self.tab_index + 1 < length:
                        self.tab_index += 1
                    else:
                        self.tab_index = 0
                else:
                    self.err_msg = f"This field does not match the minimum length of {min_len}"
            elif keypress == Keyboard.VK_BACKSPACE:  # Handle backspace
                if self.cur_field.can_write() and len(self.cur_field.value) >= 1:
                    self.cur_field.value = self.cur_field.value[:-1]
                elif "read-only" == self.cur_field.access_type:
                    self.err_msg = "This field is read only !"
            elif keypress.isalnum() or keypress.ispunct() or keypress == Keyboard.VK_SPACE:  # Handle an input
                if not self.cur_field.can_write():
                    self.err_msg = "This field is read only !"
                else:
                    if self.cur_field.mode == "checkbox":
                        if keypress == Keyboard.VK_SPACE:
                            if not self.cur_field.value:
                                self.cur_field.value = 1
                            else:
                                self.cur_field.value = 0
                    else:
                        if len(str(self.cur_field.value)) < self.cur_field.max_length:
                            if re.match(self.cur_field.val_regex(1, self.cur_field.max_length), keypress.value):
                                # Append value to the current field if the value matches the input type
                                self.cur_field.value += keypress.value
                            else:
                                self.err_msg = f"This {self.cur_field.mode} field only accept {self.cur_field.kind}s !"
            elif keypress == Keyboard.VK_UP:  # Cursor up
                if self.tab_index - 1 >= 0:
                    self.tab_index -= 1
            elif keypress == Keyboard.VK_DOWN:  # Cursor down
                if self.tab_index + 1 < length:
                    self.tab_index += 1
            elif keypress == Keyboard.VK_ENTER:  # Validate & Save the form and exit
                for idx in range(0, length):
                    field = self.all_fields[idx]
                    val_regex = field.val_regex(field.min_length, field.max_length)
                    if not re.match(val_regex, str(field.value)):
                        self.err_msg = f"Field \"{field.label}\" is not valid => \"{val_regex}\" !"
                        keypress = None
                        self.__display_error__()
                        break

        return keypress



import os
import re
from typing import Optional, List


class Properties:
    _default_name = 'application.properties'
    _profiled_format = 'application-{}.properties'

    def __init__(self,
                 filename: str = None,
                 profile: str = None,
                 load_dir: str = None):
        self.filename = filename
        self.profile = profile if profile else os.environ.get('ACTIVE_PROFILE')
        self.load_dir = load_dir if load_dir else os.path.abspath(os.curdir)
        self.properties = {}
        self._read()

    def __str__(self):
        str_val = ''
        for key, value in self.properties.items():
            str_val += '{}{}={}'.format('\n' if str_val else '', key, value)
        return str_val

    def get(self, key: str, default_value=None) -> Optional[str]:
        return self.properties[key.strip()] if key.strip() in self.properties else default_value

    def get_int(self, key: str, default_value=None) -> Optional[int]:
        try:
            return int(self.get(key))
        except TypeError:
            return default_value

    def get_float(self, key: str, default_value=None) -> Optional[float]:
        try:
            return float(self.get(key))
        except TypeError:
            return default_value

    def get_bool(self, key: str, default_value=None) -> Optional[bool]:
        return self.get(key).capitalize() in ['True', '1', 'on', 'yes']

    def size(self) -> int:
        return len(self.properties) if self.properties else 0

    def _read(self) -> None:
        default_filename = Properties._profiled_format.format(
            self.profile) if self.profile else Properties._default_name
        self.filename = self.filename if self.filename else default_filename
        file_path = '{}/{}'.format(self.load_dir, self.filename)
        if os.path.exists(file_path):
            with open(file_path) as f_properties:
                all_properties = list(map(str.strip, filter(None, f_properties.readlines())))
                self._parse(all_properties)
        else:
            raise OSError('{}: File "{}" does not exist'.format(self.__class__.__name__, file_path))

    def _parse(self, all_properties: List[str]) -> None:
        self.properties = {
            p[0].strip(): p[1].strip() for p in [
                p.split('=', 1) for p in list(
                    filter(lambda l: re.match('[a-zA-Z0-9][._\\-a-zA-Z0-9]* *=.*', l), all_properties)
                )
            ]
        }



from abc import abstractmethod

from PyQt5 import uic

from hspylib.ui.qt.views.qt_view import QtView


class MainView(QtView):

    def __init__(self, ui_file_path: str):
        form, window = uic.loadUiType(ui_file_path)
        super().__init__(window)
        self.form = form
        self.form.setupUi(window)
        self.__setup_ui()

    @abstractmethod
    def __setup_ui(self) -> None:
        pass

    def show(self) -> None:
        self.window.show()



from typing import Tuple, List, Any

from hspylib.core.tools.validator import Validator


class ArgumentValidator(Validator):

    def __call__(self, *args, **kwargs) -> Tuple[bool, List[dict]]:
        pass

    @staticmethod
    def check_arguments(arguments: Tuple[Any], req_args_num: int) -> Any:
        assert len(arguments) >= req_args_num, \
            "Invalid number of arguments = {}, expecting >= {}".format(
                len(arguments) if arguments else '0', req_args_num)

        return arguments



import pathlib

import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# The version of the package
VERSION = (HERE / "hspylib/.version").read_text()

# The package requirements
REQUIREMENTS = list(filter(None, (HERE / "requirements.txt").read_text().splitlines()))

# This call to setup() does all the work
setuptools.setup(
    name='hspylib',
    version=VERSION,
    description='HomeSetup python library',
    author='Hugo Saporetti Junior',
    author_email='yorevs@hotmail.com',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/yorevs/hspylib",
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=REQUIREMENTS,
)



#!/usr/bin/env python3
import signal
from typing import Any, Callable, List

from hspylib.core.tools.commons import sysout, set_enable_echo
from hspylib.core.tools.keyboard import Keyboard
from hspylib.ui.cli.icons.font_awesome.awesome import Awesome
from hspylib.ui.cli.menu.menu_utils import MenuUtils
from hspylib.ui.cli.vt100.vt_100 import Vt100
from hspylib.ui.cli.vt100.vt_codes import vt_print
from hspylib.ui.cli.vt100.vt_colors import VtColors


def mdashboard(
        items: List[Any],
        items_per_line: int = 5,
        title: str = 'Please select one item',
        title_color: VtColors = VtColors.ORANGE,
        nav_color: VtColors = VtColors.YELLOW) -> Any:

    return MenuDashBoard(items, items_per_line).show(title, title_color, nav_color)


class MenuDashBoard:

    CELL_TPL = [
        [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
        [' ', ' ', ' ', 'X', ' ', ' ', ' ', ' '],
        [' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ']
    ]

    SEL_CELL_TPL = [
        [' ', '\u250F', '\u2501', ' ', ' ', '\u2501', '\u2513', ' '],
        [' ', ' ', ' ', 'X', ' ', ' ', ' ', ' '],
        [' ', '\u2517', '\u2501', ' ', ' ', '\u2501', '\u251B', ' ']
    ]

    @staticmethod
    class DashBoardItem:
        def __init__(
                self,
                icon: Awesome = None,
                tooltip: str = None,
                action: Callable = None):

            self.icon = icon
            self.tooltip = tooltip
            self.action = action

    @staticmethod
    class DashBoardBuilder:
        def __init__(self):
            self.items = []

        def item(self) -> Any:
            return MenuDashBoard.ItemBuilder(self)

        def build(self) -> list:
            return self.items

    @staticmethod
    class ItemBuilder:
        def __init__(self, parent: Any):
            self.parent = parent
            self.item = MenuDashBoard.DashBoardItem()

        def icon(self, icon: Awesome):
            self.item.icon = icon
            return self

        def tooltip(self, tooltip: str):
            self.item.tooltip = tooltip
            return self

        def action(self, action: Callable):
            self.item.action = action
            return self

        def build(self):
            self.parent.items.append(self.item)
            return self.parent

    @classmethod
    def builder(cls):
        return cls.DashBoardBuilder()

    def __init__(
            self,
            items: List[Any],
            items_per_line: int = 5):

        self.all_items = items
        self.done = None
        self.re_render = True
        self.tab_index = 0
        self.items_per_line = items_per_line

    def show(
            self,
            title: str = 'Please select one item',
            title_color: VtColors = VtColors.ORANGE,
            nav_color: VtColors = VtColors.YELLOW
    ):
        ret_val = None
        length = len(self.all_items)
        signal.signal(signal.SIGINT, MenuUtils.exit_app)
        signal.signal(signal.SIGHUP, MenuUtils.exit_app)

        if length > 0:
            sysout(f"%ED2%%HOM%{title_color.placeholder()}{title}")
            vt_print(Vt100.set_auto_wrap(False))
            vt_print('%HOM%%CUD(1)%%ED0%')
            vt_print(Vt100.save_cursor())

            # Wait for user interaction
            while not self.done and ret_val != Keyboard.VK_ENTER and ret_val != Keyboard.VK_ESC:
                # Menu Renderization {
                if self.re_render:
                    self.__render__(nav_color)
                # } Menu Renderization

                # Navigation input {
                ret_val = self.__nav_input__()
                self.re_render = True
                # } Navigation input

        vt_print('%HOM%%ED2%%MOD(0)%')
        vt_print(Vt100.set_show_cursor(True))

        selected = self.all_items[self.tab_index] if ret_val == Keyboard.VK_ENTER else None
        if selected and selected.action:
            selected.action()

        return selected

    def __render__(self, nav_color: VtColors):

        vt_print(Vt100.set_show_cursor(False))
        # Restore the cursor to the home position
        vt_print(Vt100.restore_cursor())
        sysout('%NC%')
        set_enable_echo()
        # Print cells
        for idx, item in enumerate(self.all_items):
            if self.tab_index != idx:
                self.__print_cell__(idx, item, MenuDashBoard.CELL_TPL)
            else:
                self.__print_cell__(idx, item, MenuDashBoard.SEL_CELL_TPL)
        # Print selected tab tooltip
        sysout(f'\r%EL2%> %GREEN%{self.all_items[self.tab_index].tooltip}%NC%\n\n')
        sysout(f"{nav_color.placeholder()}[Enter] Select  [\u2190\u2191\u2192\u2193] Navigate  [Tab] Next  [Esc] Quit %EL0%", end='')

    def __print_cell__(self, idx: int, item: DashBoardItem, cell_template: List[List[str]]):
        num_cols = len(cell_template[0])
        num_rows = len(cell_template)
        for row in range(0, num_rows):
            for col in range(0, num_cols):
                if 'X' == cell_template[row][col]:
                    vt_print(f'{item.icon}')
                else:
                    vt_print(f'{cell_template[row][col]}')
            vt_print(f'%CUD(1)%%CUB({num_cols})%')
        if idx > 0 and (idx+1) % self.items_per_line == 0:
            # Break the line
            vt_print(f'%CUD(1)%%CUB({num_cols*self.items_per_line})%')
        elif idx + 1 < len(self.all_items):
            # Same line
            vt_print(f'%CUU({num_rows})%%CUF({num_cols})%')
        else:
            vt_print('%CUD(1)%%EL2%')

    def __nav_input__(self):
        length = len(self.all_items)
        keypress = Keyboard.read_keystroke()

        if not keypress:
            return None
        elif keypress == Keyboard.VK_ESC:
            self.done = True
            sysout('\n%NC%')
        else:
            if keypress == Keyboard.VK_TAB:  # Handle TAB
                if self.tab_index + 1 < length:
                    self.tab_index += 1
                else:
                    self.tab_index = 0
            elif keypress == Keyboard.VK_SHIFT_TAB:  # Handle Shift + TAB
                if self.tab_index - 1 >= 0:
                    self.tab_index -= 1
                else:
                    self.tab_index = length - 1
            elif keypress == Keyboard.VK_UP:  # Cursor up
                self.tab_index = max(0, self.tab_index - self.items_per_line)
            elif keypress == Keyboard.VK_DOWN:  # Cursor down
                self.tab_index = min(length - 1, self.tab_index + self.items_per_line)
            elif keypress == Keyboard.VK_LEFT:  # Cursor left
                self.tab_index = max(0, self.tab_index - 1)
            elif keypress == Keyboard.VK_RIGHT:  # Cursor right
                self.tab_index = min(length - 1, self.tab_index + 1)
            elif keypress == Keyboard.VK_ENTER:  # Select and exit
                pass

        return keypress



from typing import Optional, List, TypeVar, Generic
from uuid import UUID

from hspylib.core.exception.not_found_error import NotFoundError

ET = TypeVar('ET')
RT = TypeVar('RT')


class CrudService(Generic[ET, RT]):

    def __init__(self, repository: RT):
        self.repository = repository

    def save(self, entity: ET) -> None:
        if not self.get(entity.uuid):
            self.repository.insert(entity)
        else:
            self.repository.update(entity)

    def remove(self, entity: ET) -> None:
        if not self.get(entity.uuid):
            raise NotFoundError(
                "{} was not found: {}".format(entity.__class__, entity))
        else:
            self.repository.delete(entity)

    def list(self, filters: str = None) -> List[ET]:
        ret_val = self.repository.find_all(filters)
        return ret_val if ret_val else []

    def get(self, uuid: UUID) -> Optional[ET]:
        return self.repository.find_by_id(str(uuid))



from abc import ABC, abstractmethod
from typing import Any


class Menu(ABC):
    @abstractmethod
    def execute(self) -> Any:
        """
        Execute the current menu actions.
        :return: The next menu action to proceed after processing.
        """
        pass

    @abstractmethod
    def trigger_menu_item(self) -> Any:
        """
        Trigger the option action selected by the user.
        :return: The next menu action to trigger.
        """
        pass

    @abstractmethod
    def is_valid_option(self) -> bool:
        """
        Checks if the selected option is within the available menu_options.
        :return: whether the option is valid or not.
        """
        pass



import logging as log
from typing import Optional

from requests.structures import CaseInsensitiveDict

from hspylib.core.enum.http_method import HttpMethod
from hspylib.modules.fetch.fetch import fetch
from hspylib.modules.fetch.http_response import HttpResponse


def rest_call(
        url: str,
        method: HttpMethod,
        data: str,
        headers: Optional[CaseInsensitiveDict]) -> Optional[HttpResponse]:
    all_headers = {} if not headers else headers
    all_headers.update({
        "Content-Type": "text/json",
        "Accept": "*/*"
    })
    log.info('Processing REST {} {} -> {}'.format(all_headers, method, url))
    response = fetch(url=url, method=method, headers=all_headers, body=data)
    log.info('Response <=  Status: {}  Payload: {}'.format(
        response.status_code, response))

    return response



from typing import Callable


class Option:

    def __init__(
            self,
            shortopt: chr,
            longopt: str,
            has_argument: bool = False,
            opt_handler: Callable = None,
            required: bool = False):
        self.shortopt = f"-{shortopt.replace('-', '')}{':' if has_argument > 0 else ''}"
        self.longopt = f"--{longopt.replace('-', '')}{'=' if has_argument > 0 else ''}"
        self.has_argument = has_argument
        self.handler = opt_handler
        self.required = required

    def __str__(self) -> str:
        return ("shortopt: {}, longopt: {}, has_argument: {}, handler: {}".format(
            self.shortopt, self.longopt, self.has_argument, self.handler))

    def __repr__(self):
        return f"[{self.shortopt.replace('-', '').replace(':', '')}, {self.longopt.replace('-', '').replace('=', '')}]"

    def is_eq(self, opt: str):
        clean_opt = opt.replace('-', '').replace(':', '').replace('=', '').strip()
        return \
            clean_opt == self.shortopt.replace('-', '').replace(':', '').strip() or \
            clean_opt == self.longopt.replace('-', '').replace('=', '').strip()



import string
import sys
from typing import Any, Optional

import getkey

from hspylib.core.enum.enumeration import Enumeration
from hspylib.core.tools.commons import syserr

assert sys.stdin.isatty(), 'This module requires a terminal (TTY)'


class Keyboard(Enumeration):
    # Control keys
    VK_NONE = ''
    VK_ESC = getkey.keys.ESC
    VK_UP = getkey.keys.UP
    VK_DOWN = getkey.keys.DOWN
    VK_LEFT = getkey.keys.LEFT
    VK_RIGHT = getkey.keys.RIGHT
    VK_BACKSPACE = getkey.keys.BACKSPACE
    VK_INSERT = getkey.keys.INSERT
    VK_DELETE = getkey.keys.DELETE
    VK_ENTER = getkey.keys.ENTER
    VK_TAB = '\t'
    VK_SPACE = getkey.keys.SPACE
    VK_HOME = getkey.keys.HOME
    VK_END = getkey.keys.END
    VK_PAGE_UP = getkey.keys.PAGE_UP
    VK_PAGE_DOWN = getkey.keys.PAGE_DOWN
    VK_SHIFT_TAB = '\x1b[Z'
    # Letters
    VK_a = 'a'; VK_A = 'A'
    VK_b = 'b'; VK_B = 'B'
    VK_c = 'c'; VK_C = 'C'
    VK_d = 'd'; VK_D = 'D'
    VK_e = 'e'; VK_E = 'E'
    VK_f = 'f'; VK_F = 'F'
    VK_g = 'g'; VK_G = 'G'
    VK_h = 'h'; VK_H = 'H'
    VK_i = 'i'; VK_I = 'I'
    VK_j = 'j'; VK_J = 'J'
    VK_k = 'k'; VK_K = 'K'
    VK_l = 'l'; VK_L = 'L'
    VK_m = 'm'; VK_M = 'M'
    VK_n = 'n'; VK_N = 'N'
    VK_o = 'o'; VK_O = 'O'
    VK_p = 'p'; VK_P = 'P'
    VK_q = 'q'; VK_Q = 'Q'
    VK_r = 'r'; VK_R = 'R'
    VK_s = 's'; VK_S = 'S'
    VK_t = 't'; VK_T = 'T'
    VK_u = 'u'; VK_U = 'U'
    VK_v = 'v'; VK_V = 'V'
    VK_w = 'w'; VK_W = 'W'
    VK_x = 'x'; VK_X = 'X'
    VK_y = 'y'; VK_Y = 'Y'
    VK_z = 'z'; VK_Z = 'Z'
    # Numbers
    VK_ZERO = '0'
    VK_ONE = '1'
    VK_TWO = '2'
    VK_THREE = '3'
    VK_FOUR = '4'
    VK_FIVE = '5'
    VK_SIX = '6'
    VK_SEVEN = '7'
    VK_EIGHT = '8'
    VK_NINE = '9'
    # Punctuation
    VK_BACKTICK = '`'
    VK_TILDE = '~'
    VK_EXCLAMATION_MARK = '!'
    VK_AT = '@'
    VK_SHARP = '#'
    VK_DOLLAR = '$'
    VK_PERCENTAGE = '%'
    VK_CIRCUMFLEX = '^'
    VK_AMPERSAND = '&'
    VK_ASTERISK = '*'
    VK_OPEN_PARENTHESIS = '('
    VK_CLOSE_PARENTHESIS = ')'
    VK_MINUS = '-'
    VK_UNDERSCORE = '_'
    VK_PLUS = '+'
    VK_EQUALS = '='
    VK_LEFT_BRACKET = '['
    VK_RIGHT_BRACKET = ']'
    VK_LEFT_BRACE = '{'
    VK_RIGHT_BRACE = '}'
    VK_PIPE = '|'
    VK_BACK_SLASH = '\\'
    VK_BACK_COLON = ':'
    VK_BACK_SEMI_COLON = ';'
    VK_QUOTE = '\''
    VK_DOUBLE_QUOTE = '"'
    VK_LOWER = '<'
    VK_GREATER = '>'
    VK_COMMA = ','
    VK_PERIOD = '.'
    VK_SLASH = '/'
    VK_QUESTION_MARK = '?'

    @classmethod
    def read_keystroke(cls) -> Optional[Any]:
        keystroke = None
        try:
            keystroke = getkey.getkey()
            if keystroke:
                return cls.of_value(keystroke)
            else:
                return None
        except KeyboardInterrupt as err:
            syserr(str(err))
        except AssertionError:
            pass

        return None

    def isdigit(self):
        return str(self.value).isdigit()

    def isalpha(self):
        return str(self.value).isalpha()

    def isalnum(self):
        return str(self.value).isalnum()

    def ispunct(self):
        return all(ch in string.punctuation for ch in str(self.value))
